"""Guarded call"""

import hashlib
from functools import partial
from pathlib import Path

try:
    from loguru import logger as logging
except ModuleNotFoundError:
    import logging  # type: ignore

from typing import Any, Callable, Generator, Iterable

from .turbo_broccoli import load_json, save_json, to_json
from .environment import get_artifact_path, set_artifact_path
from .native import save, load


class GuardedBlockHandler:
    """
    A guarded block handler allows to guard an entire block/loop of code.

    ### Guarded block

    Use it as follows:

    ```py
    h = GuardedBlockHandler("out/foo.json")
    for _ in h.guard():
        # This whole block will be skipped if out/foo.json exists
        # If not, don't forget to set the results:
        h.result = ...
    # In any case, the results of the block are available in h.result
    ```

    (I know the syntax isn't the prettiest. It would be more natural to use a
    `with h.guard():` syntax but python doesn't allow for context managers that
    don't yield...)

    The handler's `result` is `None` by default. If left to `None`, no output
    file is created. This allows for scenarios like

    ```py
    h = GuardedBlockHandler("out/foo.json")
    for _ in h.guard():
        ... # Guarded code
        if success:
            h.result = ...
    ```

    Note that the parent directory of the output file (in this case, `out/`)
    will be created if it does not already exist.

    It is also possible to use "native" saving/loading methods:

    ```py
    h = GuardedBlockHandler("out/foo.csv")
    for _ in h.guard():
        ...
        h.result = some_pandas_dataframe
    ```

    See `turbo_broccoli.native.save` and `turbo_broccoli.native.load`.

    ### Guarded loop

    If an iterable `l` is passed to `h.guard`, then `for _ in h.guard(l):`
    iterates over the elements of `l`, and after every iteration, an artifact
    is created using the current value of `h.result` at
    `<output_path>/<str(element)>`. For example:

    ```py
    l = [1, 2, 3]  # The elements of l must be stringable
    h = GuardedBlockHandler("out/foo")
    for x in h.guard(l):
        # At the begining of every iteration, h.result is guaranteed to be a
        # dict with key str(x) being None. The result of past iterations are
        # also in h.result
        h.result[x] = {"bar": int(x) + 1}
    # now, h.result is {1: {"bar": 2}, 2: {"bar": 3}, ...}
    ```

    creates folder `out/foo` in which there will be files `1.json`, `2.json`,
    `3.json` respectively containing `{"bar": 2}`, `{"bar": 3}`, `{"bar": 4}`.
    Of course, if any of these files existed prior to running the `for` loop
    (e.g. `out/foo/2.json`), then the corresponding iteration (in that case `x
    = 2`) is skipped.
    """

    name: str | None
    result: Any = None
    output_path: Path

    def __init__(
        self, output_path: str | Path, name: str | None = None
    ) -> None:
        self.output_path = Path(output_path)
        self.name = name

    def _guard_iter(
        self, iterable: Iterable[Any]
    ) -> Generator[Any, None, None]:
        """
        Internal implementation of a guarded loop. See
        `turbo_broccoli.guard.GuardedBlockHandler`'s documentation.
        """
        self.result = {}
        for x in iterable:
            sx = str(x)
            path = self.output_path / f"{sx}.json"
            if path.is_file():
                self.result[x] = load_json(path)
                if self.name:
                    logging.debug(
                        f"Skipped guarded iteration '{self.name}'[{sx}]"
                    )
            else:
                self.result[x] = None
                yield x
                if self.result[x] is not None:
                    path.parent.mkdir(parents=True, exist_ok=True)
                    save_json(self.result[x], path)
                    if self.name is not None:
                        logging.debug(
                            f"Saved guarded iteration '{self.name}'[{sx}] "
                            f"results to '{path}'"
                        )

    def _guard_no_iter(self) -> Generator[Any, None, None]:
        """
        Internal implementation of a guarded block. See
        `turbo_broccoli.guard.GuardedBlockHandler`'s documentation.
        """
        if self.output_path.is_file():
            self.result = load(self.output_path)
            if self.name:
                logging.debug(f"Skipped guarded block '{self.name}'")
        else:
            yield self
            if self.result is not None:
                self.output_path.parent.mkdir(parents=True, exist_ok=True)
                save(self.result, self.output_path)
                if self.name is not None:
                    logging.debug(
                        f"Saved guarded block '{self.name}' results to "
                        f"'{self.output_path}'"
                    )

    def guard(
        self, iterable: Iterable[Any] | None = None
    ) -> Generator[Any, None, None]:
        """See `turbo_broccoli.guard.GuardedBlockHandler`'s documentation"""
        self.output_path.parent.mkdir(parents=True, exist_ok=True)
        old_artifact_path = get_artifact_path()
        if iterable is None:
            set_artifact_path(self.output_path.parent)
            for x in self._guard_no_iter():
                yield x
        else:
            set_artifact_path(self.output_path)
            for x in self._guard_iter(iterable):
                yield x
        set_artifact_path(old_artifact_path)


def guarded_call(
    function: Callable[..., Any],
    path: str | Path,
    *args,
    **kwargs,
) -> Any:
    """
    Convenience function:

    ```py
    guarded_call(f, "out/result.json", *args, **kwargs)
    ```
    is equivalent to

    ```py
    _f = produces_document(f, "out/result.json")
    _f(*args, **kwargs)
    ```

    Note that the parent directory of the output file (in this case, `out/`)
    will be created if it does not exist.

    Using native saving/loading is also possible:

    ```py
    guarded_call(f, "out/result.csv", *args, **kwargs)
    ```

    (assuming `f` returns a pandas DataFrame or a pandas Series)
    """
    _f = produces_document(function, path)
    return _f(*args, **kwargs)


def produces_document(
    function: Callable[..., Any],
    path: str | Path,
    check_args: bool = False,
) -> Callable[..., Any]:
    """
    Consider an expensive function `f` that returns a TurboBroccoli/JSON-izable
    `dict`. Wrapping/decorating `f` using `produces_document` essentially saves
    the result at a specified path and when possible, loads it instead of
    calling `f`. For example:

    ```py
    _f = produces_document(f, "out/result.json")
    _f(*args, **kwargs)
    ```

    will only call `f` if the `out/result.json` does not exist, and otherwise,
    loads and returns `out/result.json`. Note that the parent directory of the
    output file (in this case, `out/`) will be created if it does not
    exist. However, if `out/result.json` exists and was produced by calling
    `_f(*args, **kwargs)`, then

    ```py
    _f(*args2, **kwargs2)
    ```

    will still return the same result. If you want to keep a different file for
    each `args`/`kwargs`, set `check_args` to `True` as in

    ```py
    _f = produces_document(f, "out/result.json")
    _f(*args, **kwargs)
    ```

    In this case, the arguments must be TurboBroccoli/JSON-izable, i.e. the
    document

    ```
    {
        "args": args,
        "kwargs": kwargs,
    }
    ```

    must be TurboBroccoli/JSON-izable. The resulting file is no longer
    `out/result.json` but rather `out/result.json/<hash>.json` where `hash` is
    the MD5 hash of the serialization of the `args`/`kwargs` document above.

    Using native saving/loading is also possible:

    ```py
    _f = produces_document(f, "out/result.csv")
    _f(*args, **kwargs)
    ```

    (assuming `f` returns a pandas DataFrame or a pandas Series)
    """

    def _wrapped(path: Path, *args, **kwargs) -> dict[str, Any]:
        path.parent.mkdir(parents=True, exist_ok=True)
        old_artifact_path = get_artifact_path()
        set_artifact_path(path.parent)
        if check_args:
            s = to_json({"args": args, "kwargs": kwargs})
            h = hashlib.md5(s.encode("utf-8")).hexdigest()
            path.mkdir(parents=True, exist_ok=True)
            path = path / f"{h}.json"
        try:
            obj = load(path)
            logging.debug(
                f"Skipped call to guarded method '{function.__name__}'"
            )
        except:  # pylint: disable=bare-except
            obj = function(*args, **kwargs)
            save(obj, path)
        finally:
            set_artifact_path(old_artifact_path)
        return obj

    path = path if isinstance(path, Path) else Path(path)
    return partial(_wrapped, path)



"""Various utilities and internal methods"""

import re
from typing import Any, Generator

from turbo_broccoli.environment import is_nodecode

try:
    from loguru import logger as logging
except ModuleNotFoundError:
    import logging  # type: ignore

_WARNED_ABOUT_SAFETENSORS = False


class DeserializationError(Exception):
    """Raised whenever something went wrong during deserialization"""


class SerializationError(Exception):
    """Raised whenever something went wrong during serialization"""


class TypeNotSupported(Exception):
    """
    `to_json` will raise that if they are fed types they cannot manage. This is
    fine, the dispatch in
    `turbo_broccoli.turbo_broccoli.TurboBroccoliEncoder.default` catches these
    and moves on to the next registered `to_json` method.
    """


class TypeIsNodecode(Exception):
    """
    `from_json` methods raise this if the type shouldn't be decoded. See
    `turbo_broccoli.environment.set_nodecode`. This is fine,
    `turbo_broccoli.turbo_broccoli.TurboBroccoliDecoder._hook` catches these
    and returns `None`.
    """


def artifacts(doc: Any) -> Generator[str, None, None]:
    """
    Lists all the artifacts names referenced by this JSON document. Obviously,
    it should have been deserialized using vanilla `json.load` or `json.loads`,
    or using turbo broccoli with adequate nodecodes.

    In reality, this method recursively traverses the document and searches for
    dicts that:

    - have a `"__version__"`, `"__type__"`, and a `"id"` field;

    - the value at `"id"` is a UUID4 or has the form `<uuid4>.<...>`, i.e.
      matches the regexp

        ```re
        ^[0-9a-f]{8}(\\-[0-9a-f]{4}){3}\\-[0-9a-f]{12}(\\..+)?$
        ```

      in which case that value is `yield`ed.

    TODO:
        Implement a smarter way
    """
    if isinstance(doc, dict):
        fields = ["__version__", "__type__", "id"]
        if all(map(lambda f: f in doc, fields)):
            v = doc["id"]
            r = re.compile(
                r"^[0-9a-f]{8}(\-[0-9a-f]{4}){3}\-[0-9a-f]{12}(\..+)?$"
            )
            if r.match(v):
                yield v
        else:
            for v in doc.values():
                for a in artifacts(v):
                    yield a
    elif isinstance(doc, list):
        for x in doc:
            for a in artifacts(x):
                yield a


def raise_if_nodecode(name: str) -> None:
    """
    If the (prefixed) type name is set to nodecode
    (`turbo_broccoli.environment.set_nodecode`), raises a
    `turbo_broccoli.utils.TypeIsNodecode` exception.
    """
    if is_nodecode(name):
        raise TypeIsNodecode(name)


def warn_about_safetensors():
    """
    If safetensors is not installed, logs a warning message. This method may be
    called multiple times, but the message will only be logged once.
    """
    global _WARNED_ABOUT_SAFETENSORS  # pylint: disable=global-statement
    if not _WARNED_ABOUT_SAFETENSORS:
        logging.warning(
            "Serialization of numpy arrays and Tensorflow tensors without "
            "safetensors is deprecated. Consider installing safetensors using "
            "'pip install safetensors'."
        )
        _WARNED_ABOUT_SAFETENSORS = True



# pylint: disable=missing-function-docstring
"""Test suite for guarded blocks/loops"""

from pathlib import Path

# pylint: disable=unused-import
# pylint: disable=import-outside-toplevel

# Must be before turbo_broccoli imports
import common

from turbo_broccoli import (
    GuardedBlockHandler,
    guarded_call,
    produces_document,
    load_json,
)

TEST_PATH = Path("out") / "test"


def test_guarded_bloc_handler_iter():
    path = TEST_PATH / "test_guarded_bloc_handler_iter"
    h = GuardedBlockHandler(path)
    l1, l2 = [1, 2, 3], [2, 3, 4]
    # First loop
    for x in h.guard(l1):
        # Initialization of results at each iteration
        assert x in h.result
        assert h.result[x] is None
        h.result[x] = int(x)
    # Second loop over same iterable should be skipped
    for x in h.guard(l1):
        assert False
    # Final value
    assert h.result == {i: i for i in l1}
    # Check output files individually
    for x in l1:
        p = path / f"{x}.json"
        assert p.is_file()
        assert load_json(p) == int(x)
    # Second iteration where some output files already exist
    for x in h.guard(l2):
        if (
            int(x) <= 3
        ):  # Iterations for files that already exist should be skipped
            assert False
        h.result[x] = int(x)
    assert h.result == {i: i for i in l2}


def test_guarded_bloc_handler_no_iter():
    path = TEST_PATH / "test_guarded_bloc_handler_no_iter.json"
    h = GuardedBlockHandler(path)
    for _ in h.guard():
        h.result = 41
        h.result = 42
    # Block should be skipped
    for _ in h.guard():
        assert False
    assert h.result == 42
    assert h.result == load_json(path)


def test_guarded_bloc_handler_no_iter_native():
    import pandas as pd

    path = TEST_PATH / "test_guarded_bloc_handler_no_iter_native.csv"
    h = GuardedBlockHandler(path)
    df1 = pd.DataFrame({"A": [1, 2, 3], "B": [1.0, 2.0, 3.0]})
    for _ in h.guard():
        h.result = df1
    for _ in h.guard():  # Block should be skipped
        assert False
    df2 = pd.read_csv(path)
    if "Unnamed: 0" in df2.columns:
        df2.drop(["Unnamed: 0"], axis=1, inplace=True)
    assert (df1 == df2).all().all()


def test_guarded_call():
    def f(a: int):
        return {"a": a}

    path = TEST_PATH / "test_guarded_call.json"
    x = guarded_call(f, path, 1)
    y = load_json(path)
    assert isinstance(x, dict)
    assert x == y
    assert x != f(2)
    assert x == guarded_call(f, path, 2)  # Intended behavior


def test_produces_document():
    def f(a: int):
        return {"a": a}

    path = TEST_PATH / "test_produces_document.json"
    _f = produces_document(f, path, check_args=False)
    x = _f(1)
    y = load_json(path)
    assert isinstance(x, dict)
    assert x == y
    assert x != f(2)
    assert x == _f(2)  # Intended behavior


def test_produces_document_check_args():
    def f(a: int):
        return {"a": a}

    path = TEST_PATH / "test_produces_document_check_args.json"
    _f = produces_document(f, path, check_args=True)
    assert _f(1) == {"a": 1}
    assert _f(2) == {"a": 2}
    assert _f(1) == {"a": 1}  # Repetition intended
    assert _f(2) == {"a": 2}  # Repetition intended
    assert _f(1) != _f(2)



# pylint: disable=missing-function-docstring
"""bytes (de)serialization test suite"""

from common import from_json, to_json


def test_bytes_empty():
    x = b""
    assert x == from_json(to_json(x))


def test_bytes_ascii():
    x = "Hello".encode("ascii")
    assert x == from_json(to_json(x))


def test_bytes_utf8():
    x = "Hello 👋".encode("utf8")
    assert x == from_json(to_json(x))



# pylint: disable=missing-function-docstring
"""deque (de)serialization test suite"""

from common import from_json, to_json


# pylint: disable=missing-class-docstring
class C:
    __turbo_broccoli__ = ["a_str", "an_int"]
    a_byte_str: bytes
    a_list: list
    a_str: str

    @property
    def an_int(self):
        return 2 * len(self.a_list)


def test_dataclass():
    x = C()
    x.a_byte_str = "🦐🦐🦐".encode("utf8")
    x.a_list = list(range(10))
    x.a_str = "Hello 🌎"
    y = {"a_str": x.a_str, "an_int": x.an_int}
    z = from_json(to_json(x))["__generic__"]["data"]
    assert y == z



# pylint: disable=missing-function-docstring
"""(de)serialization of secrets"""

import nacl.secret
import nacl.utils
import pytest
from common import from_json, to_json  # Must be before turbo_broccoli imports
from nacl.exceptions import CryptoError

from turbo_broccoli.environment import set_shared_key
from turbo_broccoli.secret import (
    LockedSecret,
    Secret,
    SecretDict,
    SecretFloat,
    SecretInt,
    SecretList,
    SecretStr,
)


def _new_key() -> bytes:
    return nacl.utils.random(nacl.secret.SecretBox.KEY_SIZE)


def assert_secret_equal(a: Secret, b: Secret):
    assert type(a) == type(b)  # pylint: disable=unidiomatic-typecheck
    if not isinstance(a, LockedSecret):
        assert a.get_secret_value() == b.get_secret_value()


def test_secret():
    set_shared_key(_new_key())
    x = {
        "a_dict": SecretDict({"a": 1, "b": 2}),
        "a_float": SecretFloat(1.2),
        "a_int": SecretInt(12),
        "a_list": SecretList(list(range(10))),
        "a_str": SecretStr("password"),
    }
    y = from_json(to_json(x))
    assert_secret_equal(x["a_dict"], y["a_dict"])
    assert_secret_equal(x["a_float"], y["a_float"])
    assert_secret_equal(x["a_int"], y["a_int"])
    assert_secret_equal(x["a_list"], y["a_list"])
    assert_secret_equal(x["a_str"], y["a_str"])
    set_shared_key(None)


def test_secret_nokey():
    set_shared_key(_new_key())
    x = {
        "a_dict": SecretDict({"a": 1, "b": 2}),
        "a_float": SecretFloat(1.2),
        "a_int": SecretInt(12),
        "a_list": SecretList(list(range(10))),
        "a_str": SecretStr("password"),
    }
    y = to_json(x)
    set_shared_key(None)
    z = from_json(y)
    assert isinstance(z["a_dict"], LockedSecret)
    assert isinstance(z["a_float"], LockedSecret)
    assert isinstance(z["a_int"], LockedSecret)
    assert isinstance(z["a_list"], LockedSecret)
    assert isinstance(z["a_str"], LockedSecret)


def test_secret_wrongkey():
    set_shared_key(_new_key())
    x = {
        "a_dict": SecretDict({"a": 1, "b": 2}),
        "a_float": SecretFloat(1.2),
        "a_int": SecretInt(12),
        "a_list": SecretList(list(range(10))),
        "a_str": SecretStr("password"),
    }
    y = to_json(x)
    set_shared_key(_new_key())
    with pytest.raises(CryptoError):
        from_json(y)
    set_shared_key(None)



# pylint: disable=missing-function-docstring
"""Tensorflow (de)serialization test suite"""

import os
import tensorflow as tf

from common import from_json, to_json


def _assert_sparse_equal(a, b):
    tf.debugging.assert_equal(a.dense_shape, b.dense_shape)
    tf.debugging.assert_equal(a.indices, b.indices)
    tf.debugging.assert_equal(a.values, b.values)


def test_tensorflow_numerical():
    x = tf.constant([])
    tf.debugging.assert_equal(x, from_json(to_json(x)))
    x = tf.constant([1, 2, 3])
    tf.debugging.assert_equal(x, from_json(to_json(x)))
    x = tf.random.uniform((10, 10))
    tf.debugging.assert_equal(x, from_json(to_json(x)))


def test_tensorflow_numerical_large():
    os.environ["TB_MAX_NBYTES"] = "0"
    x = tf.random.uniform((100, 100), dtype=tf.float64)
    tf.debugging.assert_equal(x, from_json(to_json(x)))


def test_tensorflow_sparse():
    x = tf.SparseTensor(indices=[[0, 0]], values=[1], dense_shape=(1, 1))
    _assert_sparse_equal(x, from_json(to_json(x)))
    x = tf.SparseTensor(
        indices=[[0, 3], [2, 4]],
        values=[10, 20],
        dense_shape=(10_000_000_000, 10_000_000_000),
    )
    _assert_sparse_equal(x, from_json(to_json(x)))


def test_tensorflow_variable():
    x = tf.Variable(1.0)
    tf.debugging.assert_equal(x, from_json(to_json(x)))
    x = tf.Variable([1.0])
    tf.debugging.assert_equal(x, from_json(to_json(x)))
    x = tf.Variable(tf.random.uniform((10, 10)))
    tf.debugging.assert_equal(x, from_json(to_json(x)))



# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
"""deque (de)serialization test suite"""

import numpy as np
from numpy.testing import assert_array_equal
from scipy.sparse import csr_matrix

from common import from_json, to_json


def _assert_csr_matrix_equal(m1: csr_matrix, m2: csr_matrix) -> None:
    assert not (m1 != m2).toarray().any()
    assert_array_equal(m1.toarray(), m2.toarray())


def test_csr_matrix():
    m = csr_matrix((3, 4), dtype=np.int8)
    _assert_csr_matrix_equal(m, from_json(to_json(m)))

    row = np.array([0, 0, 1, 2, 2, 2])
    col = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6])
    m = csr_matrix((data, (row, col)), shape=(3, 3))
    _assert_csr_matrix_equal(m, from_json(to_json(m)))

    indptr = np.array([0, 2, 3, 6])
    indices = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6])
    m = csr_matrix((data, indices, indptr), shape=(3, 3))
    _assert_csr_matrix_equal(m, from_json(to_json(m)))

    row = np.array([0, 1, 2, 0])
    col = np.array([0, 1, 1, 0])
    data = np.array([1, 2, 4, 8])
    m = csr_matrix((data, (row, col)), shape=(3, 3))
    _assert_csr_matrix_equal(m, from_json(to_json(m)))



"""
numpy (de)serialization utilities.

Todo:
    Handle numpy's `generic` type (which supersedes the `number` type).
"""
__docformat__ = "google"

import pickle
from typing import Any, Callable, Tuple
from uuid import uuid4

import numpy as np

try:
    from safetensors import numpy as st

    HAS_SAFETENSORS = True
except ModuleNotFoundError:
    HAS_SAFETENSORS = False
    from turbo_broccoli.utils import warn_about_safetensors

    warn_about_safetensors()


from turbo_broccoli.environment import (
    get_artifact_path,
    get_max_nbytes,
)
from turbo_broccoli.utils import (
    DeserializationError,
    TypeNotSupported,
    raise_if_nodecode,
)


def _json_to_dtype(dct: dict) -> np.dtype:
    """
    Converts a JSON document to a numpy dtype. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__numpy__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_dtype_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_dtype_v1(dct: dict) -> np.dtype:
    """
    Converts a JSON document to a numpy dtype object following the v1
    specification.
    """
    return np.lib.format.descr_to_dtype(dct["dtype"])


def _json_to_ndarray(dct: dict) -> np.ndarray:
    """
    Converts a JSON document to a numpy array. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__numpy__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_ndarray_v1,
        2: _json_to_ndarray_v2,
        3: _json_to_ndarray_v3,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_ndarray_v1(dct: dict) -> np.ndarray:
    """
    Converts a JSON document to a numpy array following the v1 specification.
    """
    return np.frombuffer(
        dct["data"],
        dtype=np.lib.format.descr_to_dtype(dct["dtype"]),
    ).reshape(dct["shape"])


def _json_to_ndarray_v2(dct: dict) -> np.ndarray:
    """
    Converts a JSON document to a numpy array following the v1 specification.
    """
    if "data" in dct:
        return _json_to_ndarray_v1(dct)
    return np.load(get_artifact_path() / (dct["id"] + ".npy"))


def _json_to_ndarray_v3(dct: dict) -> np.ndarray:
    """
    Converts a JSON document to a numpy array following the v1 specification.
    """
    if not HAS_SAFETENSORS:
        raise RuntimeError(
            "A v3 numpy array document cannot be deserialized without "
            "safetensors. Install safetensors using "
            "'pip install safetensors'"
        )
    if "data" in dct:
        return st.load(dct["data"])["data"]
    return st.load_file(get_artifact_path() / dct["id"])["data"]


def _json_to_number(dct: dict) -> np.number:
    """
    Converts a JSON document to a numpy number. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__numpy__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_number_v1,
        2: _json_to_number_v2,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_number_v1(dct: dict) -> np.number:
    """
    Converts a JSON document to a numpy number following the v1 specification.
    """
    return np.frombuffer(
        dct["value"],
        dtype=np.lib.format.descr_to_dtype(dct["dtype"]),
    )[0]


def _json_to_number_v2(dct: dict) -> np.number:
    """
    Converts a JSON document to a numpy number following the v2 specification.
    """
    return np.frombuffer(dct["value"], dtype=dct["dtype"])[0]


def _json_to_random_state(dct: dict) -> np.number:
    """
    Converts a JSON document to a numpy random state. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__numpy__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_random_state_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_random_state_v1(dct: dict) -> np.number:
    """
    Converts a JSON document to a numpy random state following the v1
    specification.
    """
    with open(get_artifact_path() / dct["data"], mode="rb") as fp:
        return pickle.load(fp)


def _dtype_to_json(d: np.dtype) -> dict:
    """Serializes a `numpy` array."""
    return {
        "__type__": "dtype",
        "__version__": 1,
        "dtype": np.lib.format.dtype_to_descr(d),
    }


def _ndarray_to_json(arr: np.ndarray) -> dict:
    """Serializes a `numpy` array."""
    return (
        _ndarray_to_json_v3(arr)
        if HAS_SAFETENSORS
        else _ndarray_to_json_v2(arr)
    )


def _ndarray_to_json_v2(arr: np.ndarray) -> dict:
    """
    Serializes a `numpy` array using the v2 format (which doesn't use
    `safetensors`)
    """
    if arr.nbytes <= get_max_nbytes():
        return {
            "__type__": "ndarray",
            "__version__": 2,
            "data": bytes(arr.data),
            "dtype": np.lib.format.dtype_to_descr(arr.dtype),
            "shape": arr.shape,
        }
    name = str(uuid4())
    np.save(get_artifact_path() / name, arr)
    return {
        "__type__": "ndarray",
        "__version__": 2,
        "id": name,
    }


def _ndarray_to_json_v3(arr: np.ndarray) -> dict:
    """
    Serializes a `numpy` array using the v3 format (which does use
    `safetensors`)
    """
    if arr.nbytes <= get_max_nbytes():
        return {
            "__type__": "ndarray",
            "__version__": 3,
            "data": st.save({"data": arr}),
        }
    name = str(uuid4())
    st.save_file({"data": arr}, get_artifact_path() / name)
    return {
        "__type__": "ndarray",
        "__version__": 3,
        "id": name,
    }


def _number_to_json(num: np.number) -> dict:
    """Serializes a `numpy` number."""

    return {
        "__type__": "number",
        "__version__": 2,
        "value": bytes(np.array(num).data),
        "dtype": num.dtype,
    }


def _random_state_to_json(obj: np.random.RandomState) -> dict:
    """Pickles a numpy random state"""
    name, protocol = str(uuid4()), pickle.HIGHEST_PROTOCOL
    with open(get_artifact_path() / name, mode="wb") as fp:
        pickle.dump(obj, fp, protocol=protocol)
    return {
        "__type__": "random_state",
        "__version__": 1,
        "data": name,
        "protocol": protocol,
    }


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a numpy object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__numpy__`.
    """
    raise_if_nodecode("numpy")
    DECODERS = {
        "ndarray": _json_to_ndarray,
        "number": _json_to_number,
        "dtype": _json_to_dtype,
        "random_state": _json_to_random_state,
    }
    try:
        type_name = dct["__numpy__"]["__type__"]
        raise_if_nodecode("numpy." + type_name)
        return DECODERS[type_name](dct["__numpy__"])
    except KeyError as exc:
        raise DeserializationError() from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a `numpy` object into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__numpy__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    - `numpy.ndarray`: An array is processed differently depending on its size
      and on the `TB_MAX_NBYTES` environment variable. If the array is
      small, i.e. `arr.nbytes <= TB_MAX_NBYTES`, then it is directly
      stored in the resulting JSON document as

            {
                "__numpy__": {
                    "__type__": "ndarray",
                    "__version__": 2,
                    "data": <ASCII encoded byte string>,
                    "dtype": <dtype_to_descr string>,
                    "shape": <int tuple>,
                }
            }

      or, if the `safetensors` package is available:

            {
                "__numpy__": {
                    "__type__": "ndarray",
                    "__version__": 3,
                    "data": <ASCII encoded byte string>,
                }
            }


      On the other hand, the array is too large (i.e. the number of bytes
      exceeds `TB_MAX_NBYTES` or the value set by
      `turbo_broccoli.environment.set_max_nbytes`), then the content of `arr`
      is stored in an `.npy` file. Said file is saved to the path specified by
      the `TB_ARTIFACT_PATH` environment variable with a
      random UUID4 as filename. The resulting JSON document looks like

            {
                "__numpy__": {
                    "__type__": "ndarray",
                    "__version__": <2 or 3>,
                    "id": <UUID4 str>,
                }
            }

      By default, `TB_MAX_NBYTES` is `8000` bytes, which should be enough
      to store an array of 1000 `float64`s, and `TB_ARTIFACT_PATH` is `./`.
      `TB_ARTIFACT_PATH` must point to an existing directory.

    - `numpy.number`:

            {
                "__numpy__": {
                    "__type__": "number",
                    "__version__": 2,
                    "value": <float>,
                    "dtype": {...},
                }
            }

        where the `dtype` document follows the specification below.

    - `numpy.dtype`:

            {
                "__numpy__": {
                    "__type__": "dtype",
                    "__version__": 1,
                    "dtype": <dtype_to_descr string>,
                }
            }

    - `numpy.random.RandomState`:

            {
                "__numpy__": {
                    "__type__": "random_state",
                    "__version__": 1,
                    "dtype": <uuid4>,
                    "protocol": <int>
                }
            }

      where the UUID4 points to a pickle file artefact, and the protocol is the
      pickle protocol.

    """
    ENCODERS: list[Tuple[type, Callable[[Any], dict]]] = [
        (np.ndarray, _ndarray_to_json),
        (np.number, _number_to_json),
        (np.dtype, _dtype_to_json),
        (np.random.RandomState, _random_state_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__numpy__": f(obj)}
    raise TypeNotSupported()



# pylint: disable=missing-function-docstring
"""bytes (de)serialization test suite"""

import json

from common import from_json, to_json


def test_str_dict():
    x = {"a": 1, "b": 2}
    assert x == from_json(to_json(x))
    assert x == json.loads(json.dumps(x))


def test_int_dict():
    x, y = {1: "a", 2: "b"}, {"1": "a", "2": "b"}
    assert x == from_json(to_json(x))
    assert y == json.loads(json.dumps(x))


def test_float_dict():
    x = {1.1: "a", 2.2: "b"}
    assert x == from_json(to_json(x))


def test_composite_dict():
    x = {1: "a", "2": "b", 1e5: "c"}
    assert x == from_json(to_json(x))



"""Bokeh objects (de)serialization utilities."""
__docformat__ = "google"

from typing import Any, Callable, Tuple

from bokeh.core.serialization import (
    Buffer,
    Deserializer,
    Serialized,
    Serializer,
)
from bokeh.models import Model
from bokeh.plotting import figure as Figure

from turbo_broccoli.utils import (
    DeserializationError,
    TypeNotSupported,
    raise_if_nodecode,
)


def _buffer_to_json(obj: Buffer) -> dict:
    """Serializes a bokeh object.into a JSON document."""
    return {
        "__type__": "buffer",
        "__version__": 1,
        "id": obj.id,
        "data": obj.to_bytes(),
    }


def _generic_to_json(obj: Figure) -> dict:
    """Serializes a bokeh object.into a JSON document."""
    s = Serializer().serialize(obj)
    return {
        "__type__": "generic",
        "__version__": 1,
        "content": s.content,
        "buffers": s.buffers,
    }


def _json_to_buffer(dct: dict) -> Buffer:
    """
    Converts a JSON document to a bokeh buffer. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__bokeh__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_buffer_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_buffer_v1(dct: dict) -> Buffer:
    """
    Converts a JSON document to a bokeh buffer following the v1 specification.
    """
    return Buffer(id=dct["id"], data=dct["data"])


def _json_to_generic(dct: dict) -> Any:
    """
    Converts a JSON document Serializes a bokeh object. See `to_json` for
    the specification `dct` is expected to follow. Note that the key
    `__bokeh__` should not be present.
    """
    DECODERS = {
        1: _json_to_generic_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_generic_v1(dct: dict) -> Any:
    """
    Converts a JSON document Serializes a bokeh object.following the v1
    specification.
    """
    c, b = dct["content"], dct["buffers"]
    return Deserializer().deserialize(Serialized(content=c, buffers=b))


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict Serializes a bokeh object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__bokeh__`.
    """
    raise_if_nodecode("bokeh")
    DECODERS = {
        "buffer": _json_to_buffer,
        "generic": _json_to_generic,
    }
    try:
        type_name = dct["__bokeh__"]["__type__"]
        raise_if_nodecode("bokeh." + type_name)
        return DECODERS[type_name](dct["__bokeh__"])
    except KeyError as exc:
        raise DeserializationError() from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a bokeh object.

    The return dict has the following structure

        {
            "__bokeh__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    - `bokeh.plotting._figure.figure` or `bokeh.models.Model`:

            {
                "__bokeh__": {
                    "__type__": "generic",
                    "__version__": 1,
                    "content": {...},
                    "buffers": [...],
                }
            }

    - `bokeh.core.serialization.Buffer`: (for internal use)

            {
                "__bokeh__": {
                    "__type__": "buffer",
                    "__version__": 1,
                    "id": <str>,
                    "data": <bytes>,
                }
            }

    """
    ENCODERS: list[Tuple[type, Callable[[Any], dict]]] = [
        (Buffer, _buffer_to_json),
        (Figure, _generic_to_json),
        (Model, _generic_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__bokeh__": f(obj)}
    raise TypeNotSupported()



"""Tensorflow (de)serialization utilities."""
__docformat__ = "google"

from typing import Any, Callable, Tuple
from uuid import uuid4

import tensorflow as tf

try:
    from safetensors import tensorflow as st

    HAS_SAFETENSORS = True
except ModuleNotFoundError:
    HAS_SAFETENSORS = False
    from turbo_broccoli.utils import warn_about_safetensors

    warn_about_safetensors()

from turbo_broccoli.environment import (
    get_artifact_path,
    get_max_nbytes,
)
from turbo_broccoli.utils import (
    DeserializationError,
    TypeNotSupported,
    raise_if_nodecode,
)


def _json_to_sparse_tensor(dct: dict) -> tf.Tensor:
    """Converts a JSON document to a tensorflow tensor."""
    DECODERS = {
        1: _json_to_sparse_tensor_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_sparse_tensor_v1(dct: dict) -> tf.Tensor:
    """
    Converts a JSON document following the v1 specification to a tensorflow
    sparse tensor.
    """
    return tf.SparseTensor(
        dense_shape=dct["shape"],
        indices=dct["indices"],
        values=dct["values"],
    )


def _json_to_tensor(dct: dict) -> tf.Tensor:
    """Converts a JSON document to a tensorflow tensor."""
    DECODERS = {
        1: _json_to_tensor_v1,
        2: _json_to_tensor_v2,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_tensor_v1(dct: dict) -> tf.Tensor:
    """
    Converts a JSON document following the v1 specification to a tensorflow
    tensor.
    """
    return tf.constant(dct["numpy"], dtype=dct["dtype"])


def _json_to_tensor_v2(dct: dict) -> tf.Tensor:
    """
    Converts a JSON document following the v1 specification to a tensorflow
    tensor.
    """
    if not HAS_SAFETENSORS:
        raise RuntimeError(
            "A v2 tensorflow tensor document cannot be deserialized without "
            "safetensors. Install safetensors using "
            "'pip install safetensors'"
        )
    if "data" in dct:
        return st.load(dct["data"])["data"]
    return st.load_file(get_artifact_path() / dct["id"])["data"]


def _json_to_variable(dct: dict) -> tf.Variable:
    """Converts a JSON document to a tensorflow variable."""
    DECODERS = {
        1: _json_to_variable_v1,
        2: _json_to_variable_v2,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_variable_v1(dct: dict) -> tf.Variable:
    """
    Converts a JSON document following the v1 specification to a tensorflow
    variable.
    """
    return tf.Variable(
        dtype=dct["dtype"],
        initial_value=dct["numpy"],
        name=dct["name"],
        trainable=dct["trainable"],
    )


def _json_to_variable_v2(dct: dict) -> tf.Variable:
    """
    Converts a JSON document following the v2 specification to a tensorflow
    variable.
    """
    return tf.Variable(
        initial_value=dct["value"],
        name=dct["name"],
        trainable=dct["trainable"],
    )


def _ragged_tensor_to_json(tens: tf.Tensor) -> dict:
    """Serializes a general tensor"""
    raise NotImplementedError(
        "Serialization of ragged tensors is not supported"
    )


def _sparse_tensor_to_json(tens: tf.SparseTensor) -> dict:
    """Serializes a sparse tensor"""
    return {
        "__type__": "sparse_tensor",
        "__version__": 1,
        "indices": tens.indices,
        "shape": list(tens.dense_shape),
        "values": tens.values,
    }


def _tensor_to_json(tens: tf.Tensor) -> dict:
    """Serializes a general tensor"""
    return (
        _tensor_to_json_v2(tens)
        if HAS_SAFETENSORS
        else _tensor_to_json_v1(tens)
    )


def _tensor_to_json_v1(tens: tf.Tensor) -> dict:
    """
    Serializes a general tensor following the v1 format, (which doesn't use
    `safetensors`)
    """
    return {
        "__type__": "tensor",
        "__version__": 1,
        "dtype": tens.dtype.name,
        "numpy": tens.numpy(),
    }


def _tensor_to_json_v2(tens: tf.Tensor) -> dict:
    """
    Serializes a general tensor following the v2 format, (which does use
    `safetensors`)
    """
    if tens.numpy().nbytes <= get_max_nbytes():
        return {
            "__type__": "tensor",
            "__version__": 2,
            "data": st.save({"data": tens}),
        }
    name = str(uuid4())
    st.save_file({"data": tens}, get_artifact_path() / name)
    return {
        "__type__": "tensor",
        "__version__": 2,
        "id": name,
    }


def _variable_to_json(var: tf.Variable) -> dict:
    """Serializes a tensorflow variable"""
    return {
        "__type__": "variable",
        "__version__": 2,
        "name": var.name,
        "value": var.value(),
        "trainable": var.trainable,
    }


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a tensorflow object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__tensorflow__`.
    """
    raise_if_nodecode("tensorflow")
    DECODERS = {
        "sparse_tensor": _json_to_sparse_tensor,
        "tensor": _json_to_tensor,
        "variable": _json_to_variable,
    }
    try:
        type_name = dct["__tensorflow__"]["__type__"]
        raise_if_nodecode("tensorflow." + type_name)
        return DECODERS[type_name](dct["__tensorflow__"])
    except KeyError as exc:
        raise DeserializationError() from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a tensorflow object into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__tensorflow__": {...}
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    - `tf.RaggedTensor`: Not supported.

    - `tf.SparseTensor`:

            {
                "__tensorflow__": {
                    "__type__": "sparse_tensor",
                    "__version__": 1,
                    "indices": {...},
                    "values": {...},
                    "shape": {...},
                }
            }

      where the first two `{...}` placeholders result in the serialization of
      `tf.Tensor` (see below).

    - other `tf.Tensor` subtypes:

            {
                "__tensorflow__": {
                    "__type__": "tensor",
                    "__version__": 1,
                    "dtype": <str>,
                    "numpy": {...},
                }
            }

      or, if the `safetensors` package is available:

            {
                "__tensorflow__": {
                    "__type__": "tensor",
                    "__version__": 2,
                    "dtype": <str>,
                    "data": {...},
                }
            }

      On the other hand, if the `safetensors` package is available, and if the
      tensor is too large (i.e. the number of bytes exceeds `TB_MAX_NBYTES` or
      the value set by `turbo_broccoli.environment.set_max_nbytes`), then the
      content of the tensor is stored in a binary artefact´. Said file is saved
      to the path specified by the `TB_ARTIFACT_PATH` environment variable with
      a random UUID4 as filename. The resulting JSON document looks like

            {
                "__tensorflow__": {
                    "__type__": "tensor",
                    "__version__": 2,
                    "id": <UUID4 str>,
                }
            }

      By default, `TB_MAX_NBYTES` is `8000` bytes, which should be enough
      to store an array of 1000 `float64`s, and `TB_ARTIFACT_PATH` is `./`.
      `TB_ARTIFACT_PATH` must point to an existing directory.

    - `tf.Variable`:

            {
                "__tensorflow__": {
                    "__type__": "tensor",
                    "__version__": 2,
                    "name": <str>,
                    "value": {...},
                    "trainable": <bool>,
                }
            }

      where `{...}` is the document produced by serializing the value tensor of
      the variable.

    """
    ENCODERS: list[Tuple[type, Callable[[Any], dict]]] = [
        (tf.RaggedTensor, _ragged_tensor_to_json),
        (tf.SparseTensor, _sparse_tensor_to_json),
        (tf.Tensor, _tensor_to_json),
        (tf.Variable, _variable_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__tensorflow__": f(obj)}
    raise TypeNotSupported()



# pylint: disable=missing-function-docstring
"""Pytorch (de)serialization test suite"""

import os
import torch

from common import from_json, to_json  # Must be before turbo_broccoli imports

from turbo_broccoli.environment import register_pytorch_module_type


class _TestModule(torch.nn.Module):
    module: torch.nn.Module

    def __init__(self):
        super(_TestModule, self).__init__()
        self.module = torch.nn.Sequential(
            torch.nn.Linear(4, 2),
            torch.nn.ReLU(),
            torch.nn.Linear(2, 1),
            torch.nn.ReLU(),
        )

    def forward(self, x):
        return self.module.forward(x)


def test_pytorch_numerical():
    x = torch.Tensor()
    assert from_json(to_json(x)).numel() == 0
    x = torch.Tensor([1, 2, 3])
    torch.testing.assert_close(x, from_json(to_json(x)))
    x = torch.rand((10, 10))
    torch.testing.assert_close(x, from_json(to_json(x)))


def test_pytorch_numerical_large():
    os.environ["TB_MAX_NBYTES"] = "0"
    x = torch.rand((100, 100), dtype=torch.float64)
    torch.testing.assert_close(x, from_json(to_json(x)))


def test_pytorch_module():
    register_pytorch_module_type(_TestModule)
    x = torch.ones(4)
    a = _TestModule()
    b = from_json(to_json(a))
    torch.testing.assert_close(a(x), b(x))



"""bytes (de)serialization utilities."""
__docformat__ = "google"

from base64 import b64decode, b64encode
from typing import Any

from turbo_broccoli.utils import (
    DeserializationError,
    TypeNotSupported,
    raise_if_nodecode,
)


def _bytes_from_json_v1(dct: dict) -> bytes:
    """
    Deserializes a dict into a bytes object following the v1 specification.
    """
    return b64decode(dct["data"])


def from_json(dct: dict) -> bytes | None:
    """
    Deserializes a dict into a bytes object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__bytes__`.
    """
    raise_if_nodecode("bytes")
    DECODERS = {
        1: _bytes_from_json_v1,
    }
    try:
        return DECODERS[dct["__bytes__"]["__version__"]](dct["__bytes__"])
    except KeyError as exc:
        raise DeserializationError() from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a Python `bytes` object into JSON using a base 64 + ASCII scheme.

    The return dict has the following structure

        {
            "__bytes__": {
                "__version__": 1,
                "data": <str>,
            },
        }

    """
    if isinstance(obj, bytes):
        return {
            "__bytes__": {
                "__version__": 1,
                "data": b64encode(obj).decode("ascii"),
            },
        }
    raise TypeNotSupported()



# pylint: disable=import-outside-toplevel

"""Turbo Broccoli CLI"""
__docformat__ = "google"


import sys
from pathlib import Path
from typing import Iterable

try:
    import click
except ModuleNotFoundError:
    print(
        "Turbo Broccoli CLI requires click to be installed. You can install "
        "it by running `pip install click`.",
        file=sys.stderr,
    )
    sys.exit(-1)


@click.group()
def main():
    """Entrypoint."""


@main.command()
@click.argument(
    "file_path",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        path_type=Path,
    ),
)
@click.option(
    "-f",
    "--full-path",
    is_flag=True,
    default=False,
    help="Prints the full path of the artifacts",
)
def list_artifacts(file_path: Path, full_path: bool, *_, **__):
    """Lists the artifacts referenced by the JSON file."""
    import json

    from .environment import get_artifact_path
    from .utils import artifacts

    with open(file_path, mode="r", encoding="utf-8") as fp:
        document = json.load(fp)
    a: Iterable = artifacts(document)
    if full_path:
        p = get_artifact_path().absolute()
        a = map(lambda x: p / x, a)
    print(*a, sep="\n")


@main.command()
@click.argument(
    "file_path",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        path_type=Path,
    ),
)
@click.option("-d", "--dry-run", is_flag=True, default=False, help="Dry run")
def rm(file_path: Path, dry_run: bool, *_, **__):
    """
    Removes a json file and all the artifacts it references. Make sure the
    environment variable TB_ARTIFACT_PATH is correctly set.
    """
    import json

    from .environment import get_artifact_path
    from .utils import artifacts

    with open(file_path, mode="r", encoding="utf-8") as fp:
        document = json.load(fp)
    p = get_artifact_path().absolute()
    files = list(map(lambda x: p / x, artifacts(document))) + [
        file_path.absolute()
    ]
    for f in files:
        try:
            if not dry_run:
                f.unlink()
            print(f)
        except FileNotFoundError:
            print(f"rm: {f}: No such artifact", file=sys.stderr)


# pylint: disable=no-value-for-parameter
if __name__ == "__main__":
    main()



"""Pytorch (de)serialization utilities."""
__docformat__ = "google"

from typing import Any, Callable, Tuple
from uuid import uuid4

import safetensors.torch as st
import torch

from turbo_broccoli.environment import (
    get_artifact_path,
    get_max_nbytes,
    get_registered_pytorch_module_type,
)
from turbo_broccoli.utils import (
    DeserializationError,
    TypeNotSupported,
    raise_if_nodecode,
)


def _json_to_module(dct: dict) -> torch.nn.Module:
    """
    Converts a JSON document to a `pytorch` module. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__pytorch__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_module_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_module_v1(dct: dict) -> torch.nn.Module:
    """
    Converts a JSON document to a `pytorch` module following the v1
    specification.
    """
    module: torch.nn.Module = get_registered_pytorch_module_type(
        dct["class"]
    )()
    module.load_state_dict(dct["state"])
    return module


def _json_to_tensor(dct: dict) -> torch.Tensor:
    """
    Converts a JSON document to a `pytorch` tensor. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__pytorch__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_tensor_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_tensor_v1(dct: dict) -> torch.Tensor:
    """
    Converts a JSON document to a `pytorch` tensor following the v1 specification.
    """
    if "data" in dct:
        if dct["data"] is None:  # empty tensor
            return torch.Tensor()
        return st.load(dct["data"])["data"]
    return st.load_file(get_artifact_path() / dct["id"])["data"]


def _module_to_json(module: torch.nn.Module) -> dict:
    """Converts a pytorch `torch.nn.Module` into a JSON document."""
    return {
        "__type__": "module",
        "__version__": 1,
        "class": module.__class__.__name__,
        "state": module.state_dict(),
    }


def _tensor_to_json(tens: torch.Tensor) -> dict:
    """Converts a tensor into a JSON document."""
    x = tens.detach().cpu().contiguous()
    if x.numel() == 0:  # empty tensor
        return {
            "__type__": "tensor",
            "__version__": 1,
            "data": None,
        }
    if x.numpy().nbytes <= get_max_nbytes():
        return {
            "__type__": "tensor",
            "__version__": 1,
            "data": st.save({"data": x}),
        }
    name = str(uuid4())
    st.save_file({"data": x}, get_artifact_path() / name)
    return {
        "__type__": "tensor",
        "__version__": 1,
        "id": name,
    }


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a `pytorch` object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__pytorch__`.
    """
    raise_if_nodecode("pytorch")
    DECODERS = {
        "tensor": _json_to_tensor,
        "module": _json_to_module,
    }
    try:
        type_name = dct["__pytorch__"]["__type__"]
        raise_if_nodecode("pytorch." + type_name)
        return DECODERS[type_name](dct["__pytorch__"])
    except KeyError as exc:
        raise DeserializationError() from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a tensor into JSON by cases. See the README for the precise list
    of supported types.

    The return dict has the following structure

        {
            "__pytorch__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    - Tensor:

            {
                "__pytorch__": {
                    "__type__": "tensor",
                    "__version__": 1,
                    "data": <bytes>,
                }
            }

      or if the underlying data is too large resulting in an artifact being
      created:

            {
                "__pytorch__": {
                    "__type__": "tensor",
                    "__version__": 1,
                    "id": <UUID4 str>,
                }
            }

    - Module:

            {
                "__pytorch__": {
                    "__type__": "module",
                    "__version__": 1,
                    "class": <class name>,
                    "state": {...},
                }
            }

    """
    ENCODERS: list[Tuple[type, Callable[[Any], dict]]] = [
        (torch.nn.Module, _module_to_json),
        (torch.Tensor, _tensor_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__pytorch__": f(obj)}
    raise TypeNotSupported()



"""
"Native" saving utilities:
* `turbo_broccoli.native.save` takes a serializable/dumpable object and
  a path, and uses the file extension to choose the correct way to save the
  object to disk;
* `turbo_broccoli.native.load` does the opposite
"""

# pylint: disable=unused-argument
# pylint: disable=import-outside-toplevel

from functools import partial
from pathlib import Path
from typing import Any, Callable

from turbo_broccoli.numpy import HAS_SAFETENSORS

from .turbo_broccoli import (
    HAS_KERAS,
    HAS_NUMPY,
    HAS_PANDAS,
    HAS_PYTORCH,
    HAS_TENSORFLOW,
    load_json,
    save_json,
)


def _is_dict_of(obj: Any, value_type: type, key_type: type = str) -> bool:
    """Returns true if `obj` is a `dict[key_type, value_type]`"""
    return (
        isinstance(obj, dict)
        and all(isinstance(k, key_type) for k in obj.keys())
        and all(isinstance(v, value_type) for v in obj.values())
    )


def _load_csv(path: str | Path, **kwargs) -> Any:
    if not HAS_PANDAS:
        _raise_package_not_installed("pandas", "csv")
    import pandas as pd

    df = pd.read_csv(path, **kwargs)
    if "Unnamed: 0" in df.columns:
        df.drop(["Unnamed: 0"], axis=1, inplace=True)
    return df


def _load_keras(path: str | Path, **kwargs) -> Any:
    if not HAS_KERAS:
        _raise_package_not_installed("keras", "keras")

    import keras

    return keras.saving.load_model(path, **kwargs)


def _load_np(path: str | Path, **kwargs) -> Any:
    if not HAS_NUMPY:
        _raise_package_not_installed("numpy", ".npy/.npz")
    import numpy as np

    return np.load(path, **kwargs)


def _load_pq(path: str | Path, **kwargs) -> Any:
    if not HAS_PANDAS:
        _raise_package_not_installed("pandas", ".parquet/.pq")
    import pandas as pd

    df = pd.read_parquet(path, **kwargs)
    return df


def _load_pt(path: str | Path, **kwargs) -> Any:
    if not HAS_PYTORCH:
        _raise_package_not_installed("torch", "pt")
    import torch

    return torch.load(path, **kwargs)


def _load_st(path: str | Path, **kwargs) -> Any:
    if not HAS_SAFETENSORS:
        _raise_package_not_installed("safetensors", ".safetensors/.st")
    from safetensors import numpy as st

    return st.load_file(path, **kwargs)


def _raise_package_not_installed(package_name: str, extension: str):
    """
    Raises a `RuntimeError` with a templated error message

    Args:
        package_name (str): e.g. "numpy"
        extension (str): e.g. "npy"
    """
    if extension[0] != ".":
        extension = "." + extension
    raise RuntimeError(
        f"Cannot create or load `{extension}` file because {package_name} is "
        f"not installed. You can install {package_name} by running "
        f"python3 -m pip install {package_name}"
    )


def _raise_wrong_type(path: str | Path, obj_needs_to_be_a: str):
    """
    Raises a `TypeError` with a templated error message

    Args:
        path (str | Path): Path where the file should have been saved
        extension (str): "pandas DataFrame or Series"
    """
    raise TypeError(
        f"Could not save object to '{path}': object needs to be a "
        + obj_needs_to_be_a
    )


def _save_csv(obj: Any, path: str | Path, **kwargs) -> None:
    if not HAS_PANDAS:
        _raise_package_not_installed("pandas", "csv")
    import pandas as pd

    if not isinstance(obj, (pd.DataFrame, pd.Series)):
        _raise_wrong_type(path, "pandas DataFrame or Series")
    obj.to_csv(path, **kwargs)


def _save_keras(obj: Any, path: str | Path, **kwargs) -> None:
    if not HAS_KERAS:
        _raise_package_not_installed("keras", "keras")

    import keras

    if not isinstance(obj, keras.Model):
        _raise_wrong_type(path, "keras model")
    keras.saving.save_model(obj, path, **kwargs)


def _save_npy(obj: Any, path: str | Path, **kwargs) -> None:
    if not HAS_NUMPY:
        _raise_package_not_installed("numpy", "npy")
    import numpy as np

    if not isinstance(obj, np.ndarray):
        _raise_wrong_type(path, "numpy array")
    np.save(str(path), obj, **kwargs)


def _save_npz(obj: Any, path: str | Path, **kwargs) -> None:
    if not HAS_NUMPY:
        _raise_package_not_installed("numpy", "npz")
    import numpy as np

    if not _is_dict_of(obj, np.ndarray):
        _raise_wrong_type(path, "dict of numpy arrays")
    np.savez(str(path), **obj, **kwargs)


def _save_pq(obj: Any, path: str | Path, **kwargs) -> None:
    if not HAS_PANDAS:
        _raise_package_not_installed("pandas", ".parquet/.pq")
    import pandas as pd

    if not isinstance(obj, pd.DataFrame):
        _raise_wrong_type(path, "pandas DataFrame")
    obj.to_parquet(path, **kwargs)


def _save_pt(obj: Any, path: str | Path, **kwargs) -> None:
    if not HAS_PYTORCH:
        _raise_package_not_installed("torch", "pt")
    import torch

    if not (isinstance(obj, torch.Tensor) or _is_dict_of(obj, torch.Tensor)):
        _raise_wrong_type(path, "torch tensor or a dict of torch tensors")
    torch.save(obj, path, **kwargs)


def _save_st(obj: Any, path: str | Path, **kwargs) -> None:
    if not HAS_SAFETENSORS:
        _raise_package_not_installed("safetensors", ".safetensors/.st")
    import safetensors

    if HAS_NUMPY:
        import numpy as np

        if _is_dict_of(obj, np.ndarray):
            safetensors.numpy.save_file(obj, str(path), **kwargs)
            return

    if HAS_TENSORFLOW:
        import tensorflow as tf

        if _is_dict_of(obj, tf.Tensor):
            safetensors.tensorflow.save_file(obj, str(path), **kwargs)
            return

    if HAS_PYTORCH:
        import torch

        if _is_dict_of(obj, torch.Tensor):
            safetensors.torch.save_file(obj, str(path), **kwargs)
            return

    raise _raise_wrong_type(
        path,
        "dict of numpy arrays, a dict of tensorflow tensors, or a dict of "
        "pytorch tensors",
    )


def load(path: str | Path, **kwargs) -> Any:
    """
    Loads an object from a file using format-specific (or "native") methods.
    See `turbo_broccoli.native.save` for the list of supported file extensions.

    Warning:
        Safetensors files (`.st` or `.safetensors`) will be loaded as (dicts
        of) numpy arrays.
    """
    extension = Path(path).suffix
    methods: dict[str, Callable[[str | Path], Any]] = {
        ".csv": _load_csv,
        ".h5": _load_keras,
        ".keras": _load_keras,
        ".npy": _load_np,
        ".npz": _load_np,
        ".parquet": _load_pq,
        ".pq": _load_pq,
        ".pt": _load_pt,
        ".st": _load_st,
        ".tf": _load_keras,
    }
    method: Callable = methods.get(extension, load_json)
    return method(path, **kwargs)


def save(obj: Any, path: str | Path, **kwargs) -> None:
    """
    Saves an object using the file extension of `path` to determine the
    serialization/dumping method:

    * `.csv`:
      [`pandas.DataFrame.to_csv`](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_csv.html)
      or
      [`pandas.Series.to_csv`](https://pandas.pydata.org/docs/reference/api/pandas.Series.to_csv.html)
    * `.h5`:
      [`tf.keras.saving.save_model`](https://www.tensorflow.org/api_docs/python/tf/keras/saving/save_model)
      with `save_format="h5"`
    * `.keras`:
      [`tf.keras.saving.save_model`](https://www.tensorflow.org/api_docs/python/tf/keras/saving/save_model)
      with `save_format="keras"`
    * `.npy`:
        [`numpy.save`](https://numpy.org/doc/stable/reference/generated/numpy.save.html)
    * `.npz`:
        [`numpy.savez`](https://numpy.org/doc/stable/reference/generated/numpy.savez.html)
    * `.pq`, `.parquet`:
      [`pandas.DataFrame.to_parquet`](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_parquet.html)
    * `.pt`:
      [`torch.save`](https://pytorch.org/docs/stable/generated/torch.save.html)
    * `.safetensors`, `.st`: (for numpy arrays, pytorch tensors and tensorflow tensors)
      [safetensors](https://huggingface.co/docs/safetensors/index)
    * `.tf`:
      [`tf.keras.saving.save_model`](https://www.tensorflow.org/api_docs/python/tf/keras/saving/save_model)
      with `save_format="tf"`
    * `.json` and anything else: just forwarded to `turbo_broccoli.save_json`

    Args:
        obj (Any):
        path (str | Path):
        kwargs: Passed to the serialization method
    """
    extension = Path(path).suffix
    methods: dict[str, Callable[[Any, str | Path], None]] = {
        ".csv": _save_csv,
        ".h5": partial(_save_keras, save_format="h5"),
        ".keras": partial(_save_keras, save_format="keras"),
        ".npy": _save_npy,
        ".npz": _save_npz,
        ".parquet": _save_pq,
        ".pq": _save_pq,
        ".pt": _save_pt,
        ".st": _save_st,
        ".tf": partial(_save_keras, save_format="tf"),
    }
    method = methods.get(extension, save_json)
    method(obj, path, **kwargs)



#!/usr/bin/env python

"""Setup script"""

import setuptools

name = "turbo_broccoli"
version = "3.0.8"

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements = f.read().split()

packages = [name] + [
    name + "." + p for p in setuptools.find_packages(where="./" + name)
]

setuptools.setup(
    author="Cédric Ho Thanh",
    author_email="altaris@users.noreply.github.com",
    description="JSON (de)serialization extensions",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    name=name,
    packages=packages,
    platforms="any",
    project_urls={
        "Issues": "https://github.com/altaris/turbo-broccoli/issues",
    },
    python_requires=">=3.10",
    url="https://github.com/altaris/turbo-broccoli",
    version=version,
)



# pylint: disable=bare-except
"""Main module containing the JSON encoder and decoder classes."""
__docformat__ = "google"

import json
from pathlib import Path
from typing import Any, Callable


import turbo_broccoli.bytes
import turbo_broccoli.collections
import turbo_broccoli.dataclass
from turbo_broccoli.environment import get_artifact_path, set_artifact_path
import turbo_broccoli.generic
import turbo_broccoli.dict
from turbo_broccoli.utils import TypeIsNodecode, TypeNotSupported

try:
    import turbo_broccoli.keras

    HAS_KERAS = True
except:
    HAS_KERAS = False

try:
    import turbo_broccoli.numpy

    HAS_NUMPY = True
except:
    HAS_NUMPY = False

try:
    import turbo_broccoli.pandas

    HAS_PANDAS = True
except:
    HAS_PANDAS = False


try:
    import turbo_broccoli.secret

    HAS_SECRET = True
except:
    HAS_SECRET = False

try:
    import turbo_broccoli.tensorflow

    HAS_TENSORFLOW = True
except:
    HAS_TENSORFLOW = False

try:
    import turbo_broccoli.pytorch

    HAS_PYTORCH = True
except:
    HAS_PYTORCH = False


try:
    import turbo_broccoli.scipy

    HAS_SCIPY = True
except:
    HAS_SCIPY = False

try:
    import turbo_broccoli.sklearn

    HAS_SKLEARN = True
except:
    HAS_SKLEARN = False


try:
    import turbo_broccoli.bokeh

    HAS_BOKEH = True
except:
    HAS_BOKEH = False


class TurboBroccoliDecoder(json.JSONDecoder):
    """
    TurboBroccoli's custom JSON decoder class. See the README for the list of
    supported types.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(object_hook=self._hook, *args, **kwargs)

    def _hook(self, dct):
        """Deserialization hook"""
        DECODERS: dict[str, Callable[[dict], Any]] = {
            "__dict__": turbo_broccoli.dict.from_json,
            "__bytes__": turbo_broccoli.bytes.from_json,
        }
        if HAS_KERAS:
            DECODERS["__keras__"] = turbo_broccoli.keras.from_json
        if HAS_NUMPY:
            DECODERS["__numpy__"] = turbo_broccoli.numpy.from_json
        if HAS_PANDAS:
            DECODERS["__pandas__"] = turbo_broccoli.pandas.from_json
        if HAS_PYTORCH:
            DECODERS["__pytorch__"] = turbo_broccoli.pytorch.from_json
        if HAS_SECRET:
            DECODERS["__secret__"] = turbo_broccoli.secret.from_json
        if HAS_TENSORFLOW:
            DECODERS["__tensorflow__"] = turbo_broccoli.tensorflow.from_json
        if HAS_SCIPY:
            DECODERS["__scipy__"] = turbo_broccoli.scipy.from_json
        if HAS_SKLEARN:
            DECODERS["__sklearn__"] = turbo_broccoli.sklearn.from_json
        if HAS_BOKEH:
            DECODERS["__bokeh__"] = turbo_broccoli.bokeh.from_json
        # Intentionally put last
        DECODERS["__collections__"] = turbo_broccoli.collections.from_json
        DECODERS["__dataclass__"] = turbo_broccoli.dataclass.from_json
        for t, f in DECODERS.items():
            if t in dct:
                try:
                    return f(dct)
                except TypeIsNodecode:
                    return None
        return dct


class TurboBroccoliEncoder(json.JSONEncoder):
    """
    TurboBroccoli's custom JSON decoder class. See the README for the list of
    supported types.
    """

    def default(self, o: Any) -> Any:
        ENCODERS: list[Callable[[Any], dict]] = [
            turbo_broccoli.bytes.to_json,
        ]
        if HAS_KERAS:
            ENCODERS.append(turbo_broccoli.keras.to_json)
        if HAS_NUMPY:
            ENCODERS.append(turbo_broccoli.numpy.to_json)
        if HAS_PANDAS:
            ENCODERS.append(turbo_broccoli.pandas.to_json)
        if HAS_PYTORCH:
            ENCODERS.append(turbo_broccoli.pytorch.to_json)
        if HAS_SECRET:
            ENCODERS.append(turbo_broccoli.secret.to_json)
        if HAS_TENSORFLOW:
            ENCODERS.append(turbo_broccoli.tensorflow.to_json)
        if HAS_SCIPY:
            ENCODERS.append(turbo_broccoli.scipy.to_json)
        if HAS_SKLEARN:
            ENCODERS.append(turbo_broccoli.sklearn.to_json)
        if HAS_BOKEH:
            ENCODERS.append(turbo_broccoli.bokeh.to_json)
        # Intentionally put last
        ENCODERS += [
            turbo_broccoli.collections.to_json,
            turbo_broccoli.dataclass.to_json,
            turbo_broccoli.generic.to_json,
        ]
        for f in ENCODERS:
            try:
                return f(o)
            except TypeNotSupported:
                pass
        return super().default(o)

    def encode(self, o: Any) -> str:
        """
        Reimplementation of encode just to treat exceptional cases that need to
        be handled before `JSONEncoder.encode`.
        """
        PRIORITY_ENCODERS: list[Callable[[Any], dict]] = [
            turbo_broccoli.dict.to_json,
            turbo_broccoli.collections.to_json,
        ]
        for f in PRIORITY_ENCODERS:
            try:
                return super().encode(f(o))
            except TypeNotSupported:
                pass
        return super().encode(o)


def from_json(doc: str) -> Any:
    """Converts a JSON document string back to a Python object"""
    return json.loads(doc, cls=TurboBroccoliDecoder)


def load_json(path: str | Path, auto_artifact_path: bool = True) -> Any:
    """
    Loads and deserializes a JSON file using Turbo Broccoli

    Args:
        path (str | Path):
        auto_artifact_path (bool): If left to `True`, set the artifact path to
            the target file's parent directory before loading. After loading,
            the previous artifact path is restored. See also see
            `turbo_broccoli.environment.set_artifact_path`.
    """
    old_artifact_path = get_artifact_path()
    if auto_artifact_path:
        set_artifact_path(Path(path).parent)
    with open(path, mode="r", encoding="utf-8") as fp:
        document = json.load(fp, cls=TurboBroccoliDecoder)
    if auto_artifact_path:
        set_artifact_path(old_artifact_path)
    return document


def save_json(
    obj: Any, path: str | Path, auto_artifact_path: bool = True
) -> None:
    """
    Serializes and saves a JSON-serializable object

    Args:
        obj (Any):
        path (str | Path):
        auto_artifact_path (bool): If left to `True`, set the artifact path to
            the target file's parent directory before saving. After saving,
            the previous artifact path is restored. See also see
            `turbo_broccoli.environment.set_artifact_path`.
    """
    old_artifact_path = get_artifact_path()
    if auto_artifact_path:
        set_artifact_path(Path(path).parent)
    with open(path, mode="w", encoding="utf-8") as fp:
        fp.write(to_json(obj))
    if auto_artifact_path:
        set_artifact_path(old_artifact_path)


def to_json(obj: Any) -> str:
    """Converts an object to JSON"""
    return json.dumps(obj, cls=TurboBroccoliEncoder)



# pylint: disable=missing-function-docstring
"""Numpy (de)serialization test suite"""

import os

import numpy as np
from numpy.testing import assert_equal, assert_array_equal

from common import from_json, to_json


def _assert_equal(a, b):
    assert isinstance(a, (np.ndarray, np.number))
    assert_equal(type(a), type(b))
    assert_equal(a.dtype, b.dtype)
    if isinstance(a, np.ndarray):
        assert_array_equal(a, b)
    else:
        assert_equal(a, b)


def test_numpy_number():
    x = np.float32(-1.2)
    _assert_equal(x, from_json(to_json(x)))
    x = np.float32(0.0000000001)
    _assert_equal(x, from_json(to_json(x)))
    x = np.float32(1e10)
    _assert_equal(x, from_json(to_json(x)))
    x = np.int64(420)
    _assert_equal(x, from_json(to_json(x)))


def test_numpy_array():
    x = np.array([])
    _assert_equal(x, from_json(to_json(x)))
    x = np.array(1, dtype="int8")
    _assert_equal(x, from_json(to_json(x)))
    x = np.array(1, dtype="float32")
    _assert_equal(x, from_json(to_json(x)))
    x = np.random.random((1, 2, 3, 4, 5))
    _assert_equal(x, from_json(to_json(x)))


def test_numpy_large_array():
    os.environ["TB_MAX_NBYTES"] = "0"
    x = np.random.random((100, 100))
    _assert_equal(x, from_json(to_json(x)))


def test_numpy_dtype():
    dt = np.dtype(np.float64)
    assert dt == from_json(to_json(dt))
    dt = np.dtype(int)
    assert dt == from_json(to_json(dt))
    dt = np.dtype(object)
    assert dt == from_json(to_json(dt))
    dt = np.dtype("b")
    assert dt == from_json(to_json(dt))


def test_numpy_random_state():
    s1 = np.random.RandomState(seed=0)
    s2 = from_json(to_json(s1))
    assert s1.rand() == s2.rand()
    assert s1.rand() == s2.rand()  # twice on purpose



"""Serialization of dicts with non string keys"""
__docformat__ = "google"

from typing import Any

from turbo_broccoli.utils import (
    DeserializationError,
    TypeNotSupported,
    raise_if_nodecode,
)


def _json_to_dict_v1(dct: dict) -> dict:
    """
    Converts a JSON document representing a dict with non-string keys to a dict
    following the v1 specification. In particular, note that `dct` must not
    contain the key `__dict__`.
    """
    return {item["key"]: item["val"] for item in dct["items"]}


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a non-string key dict. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__dict__`.
    """
    raise_if_nodecode("dict")
    DECODERS = {
        1: _json_to_dict_v1,
    }
    try:
        return DECODERS[dct["__dict__"]["__version__"]](dct["__dict__"])
    except KeyError as exc:
        raise DeserializationError() from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a dict with non-string keys into JSON. The return dict has the
    following structure

        {
            "__dict__": {
                "__version__": 1,
                "items": [
                    {"key": {...}, "val": {...}},
                    ...
                ]
            },
        }

    """
    if not isinstance(obj, dict):
        raise TypeNotSupported()
    if all(map(lambda x: isinstance(x, str), obj.keys())):
        raise TypeNotSupported("All keys are strings")
    return {
        "__dict__": {
            "__version__": 1,
            "items": [{"key": key, "val": val} for key, val in obj.items()],
        },
    }
